#ifndef TLX
#define TLX
struct my_list_head {
	struct my_list_head *next, *prev;
};

enum my_zone_watermarks {
	MY_WMARK_MIN,
	MY_WMARK_LOW,
	MY_WMARK_HIGH,
	MY_NR_WMARK
};

#define MIGRATE_TYPES         5
#define MAX_ORDER 11
#define MIGRATE_PCPTYPES      3 /* the number of types on the pcp lists */
#define MAX_ZONELISTS 1
#define MAX_NUMNODES	1

struct my_free_area {
	struct my_list_head	free_list[MIGRATE_TYPES];
	unsigned long		nr_free;
};

#define MY_NR_LRU_LISTS 5

struct pglist_data;

typedef struct {
	long long counter;
} my_atomic64_t;

typedef my_atomic64_t my_atomic_long_t;

enum my_zone_stat_item {
	/* First 128 byte cacheline (assuming 64 bit words) */
	MY_NR_FREE_PAGES,
	MY_NR_LRU_BASE,
	MY_NR_INACTIVE_ANON = MY_NR_LRU_BASE, /* must match order of LRU_[IN]ACTIVE */
	MY_NR_ACTIVE_ANON,		/*  "     "     "   "       "         */
	MY_NR_INACTIVE_FILE,	/*  "     "     "   "       "         */
	MY_NR_ACTIVE_FILE,		/*  "     "     "   "       "         */
	MY_NR_UNEVICTABLE,		/*  "     "     "   "       "         */
	MY_NR_MLOCK,		/* mlock()ed pages found and moved off LRU */
	MY_NR_ANON_PAGES,	/* Mapped anonymous pages */
	MY_NR_FILE_MAPPED,	/* pagecache pages mapped into pagetables.
			   only modified from process context */
	MY_NR_FILE_PAGES,
	MY_NR_FILE_DIRTY,
	MY_NR_WRITEBACK,
	MY_NR_SLAB_RECLAIMABLE,
	MY_NR_SLAB_UNRECLAIMABLE,
	MY_NR_PAGETABLE,		/* used for pagetables */
	MY_NR_KERNEL_STACK,
	/* Second 128 byte cacheline */
	MY_NR_UNSTABLE_NFS,	/* NFS unstable pages */
	MY_NR_BOUNCE,
	MY_NR_VMSCAN_WRITE,
	MY_NR_WRITEBACK_TEMP,	/* Writeback using temporary buffers */
	MY_NR_ISOLATED_ANON,	/* Temporary isolated pages from anon lru */
	MY_NR_ISOLATED_FILE,	/* Temporary isolated pages from file lru */
	MY_NR_SHMEM,		/* shmem pages (included tmpfs/GEM pages) */
#ifdef CONFIG_NUMA
	MY_NUMA_HIT,		/* allocated in intended node */
	MY_NUMA_MISS,		/* allocated in non intended node */
	MY_NUMA_FOREIGN,		/* was intended here, hit elsewhere */
	MY_NUMA_INTERLEAVE_HIT,	/* interleaver preferred this zone */
	MY_NUMA_LOCAL,		/* allocation from local node */
	MY_NUMA_OTHER,		/* allocation from other node */
#endif
	MY_NR_VM_ZONE_STAT_ITEMS };


#define NR_CPUS		1
#define MAX_NR_ZONES 2 /* __MAX_NR_ZONES	@ */

struct my_per_cpu_pages {
	int count;		/* number of pages in the list */
	int high;		/* high watermark, emptying needed */
	int batch;		/* chunk size for buddy add/remove */

	/* Lists of pages, one per migrate type stored on the pcp-lists */
	struct my_list_head lists[MIGRATE_PCPTYPES];
};

struct my_per_cpu_pageset {
	struct my_per_cpu_pages pcp;
} ____cacheline_aligned_in_smp;

struct my_zone {
	unsigned long 		watermark[MY_NR_WMARK];
	unsigned long		lowmem_reserve[MAX_NR_ZONES];

	struct my_per_cpu_pageset	pageset[NR_CPUS];

//	spinlock_t		lock;
	struct my_free_area	free_area[MAX_ORDER];
	unsigned long		*pageblock_flags;


//	ZONE_PADDING(_pad1_)

	/* Fields commonly accessed by the page reclaim scanner */
//	spinlock_t		lru_lock;	

	struct my_zone_lru {
		struct my_list_head list;
	} lru[MY_NR_LRU_LISTS];

//	struct zone_reclaim_stat reclaim_stat;

	unsigned long		pages_scanned;	   /* since last reclaim */
	unsigned long		flags;		   /* zone flags, see below */

	/* Zone statistics */
	my_atomic_long_t		vm_stat[MY_NR_VM_ZONE_STAT_ITEMS];

	int prev_priority;
	unsigned int inactive_ratio;


//	ZONE_PADDING(_pad2_)    
//	wait_queue_head_t	* wait_table;
	unsigned long		wait_table_hash_nr_entries;
	unsigned long		wait_table_bits;
	struct pglist_data	*zone_pgdat;
	unsigned long		zone_start_pfn;
	unsigned long		spanned_pages;	/* total size, including holes */
	unsigned long		present_pages;	/* amount of memory (excluding holes) */
	const char		*name;
} ____cacheline_internodealigned_in_smp;

struct my_zoneref {
	struct my_zone *zone;	/* Pointer to actual zone */
	int zone_idx;		/* zone_idx(zoneref->zone) */
};

#define MAX_ZONES_PER_ZONELIST (MAX_NUMNODES * MAX_NR_ZONES)

struct my_zonelist {
//	struct zonelist_cache *zlcache_ptr;		     // NULL or &zlcache
	struct my_zoneref _zonerefs[MAX_ZONES_PER_ZONELIST + 1];
//#ifdef CONFIG_NUMA
//	struct zonelist_cache zlcache;			     // optional ...
//#endif
};

#define pgoff_t unsigned long
typedef unsigned short my_u16;

typedef struct {
	volatile int counter;
} my_atomic_t;

struct my_page {
	unsigned long flags;		/* Atomic flags, some possibly
					 * updated asynchronously */
	my_atomic_t _count;		/* Usage count, see below. */
	union {
		my_atomic_t _mapcount;	/* Count of ptes mapped in mms,
					 * to show when page is mapped
					 * & limit reverse map searches.
					 */
		struct {		/* SLUB */
			my_u16 inuse;
			my_u16 objects;
		};
	};
	union {
	    struct {
		unsigned long private;		/* Mapping-private opaque data:
					 	 * usually used for buffer_heads
						 * if PagePrivate set; used for
						 * swp_entry_t if PageSwapCache;
						 * indicates order in the buddy
						 * system if PG_buddy is set.
						 */
		struct address_space *mapping;	/* If low bit clear, points to
						 * inode address_space, or NULL.
						 * If page mapped as anonymous
						 * memory, low bit is set, and
						 * it points to anon_vma object:
						 * see PAGE_MAPPING_ANON below.
						 */
	    };
	    struct kmem_cache *slab;	/* SLUB: Pointer to slab */
	    struct my_page *first_page;	/* Compound tail pages */
	};
	union {
		pgoff_t index;		/* Our offset within mapping. */
		void *freelist;		/* SLUB: freelist req. slab lock */
	};
	struct my_list_head lru;		/* Pageout list, eg. active_list*/
};

typedef struct pglist_data {
	struct my_zone node_zones[MAX_NR_ZONES];
	struct my_zonelist node_zonelists[MAX_ZONELISTS];
	int nr_zones;
//#ifdef CONFIG_FLAT_NODE_MEM_MAP	/* means !SPARSEMEM */
	struct page *node_mem_map;
#ifdef CONFIG_CGROUP_MEM_RES_CTLR
	struct page_cgroup *node_page_cgroup;
#endif
//#endif
	struct bootmem_data *bdata;
#ifdef CONFIG_MEMORY_HOTPLUG
	/*
	 * Must be held any time you expect node_start_pfn, node_present_pages
	 * or node_spanned_pages stay constant.  Holding this will also
	 * guarantee that any pfn_valid() stays that way.
	 *
	 * Nests above zone->lock and zone->size_seqlock.
	 */
	spinlock_t node_size_lock;
#endif
	unsigned long node_start_pfn;
	unsigned long node_present_pages; /* total number of physical pages */
	unsigned long node_spanned_pages; /* total size of physical page
					     range, including holes */
	int node_id;
//	wait_queue_head_t kswapd_wait;
	struct task_struct *kswapd;
	int kswapd_max_order;
} pg_data_t;

#endif